"""
.. module:: MTL_Logistic_L21
   :synopsis: MTL binary logistic regression
.. moduleauthor:: Max J <https://github.com/DaPraxis>
"""
import numpy as np
from .init_opts import init_opts
from numpy import linalg as LA
from tqdm import tqdm
from tqdm import trange
import sys
import time

class MTL_Logistic_L21:
	"""MTL binary logistic regression with hinge loss and L21 penalty
	"""
	def __init__(self, opts, rho1=0.01):
		"""Initialization of MTL binary classification function

        Args:
            opts (opts): initalization class from opts
            rho1 (int, optional): L2,1-norm group Lasso parameter. Defaults to 0.01
        """
		self.opts = init_opts(opts)
		self.rho1 = rho1
		self.rho_L2 = 0
		if hasattr(opts, 'rho_L2'):
			rho_L2 = opts.rho_L2

	def fit(self, X, Y, **kwargs):
		"""Fit with training samples and train
  
        t: task number
        
        n: number of entries
        
        d: data dimension

		Args:
			X ([np.array(np.array)]): t x n x d.
            Y ([np.array(np.array)]): t x n x 1.
		"""
		if 'rho' in kwargs.keys():
			print(kwargs)
			self.rho1 = kwargs['rho']
		X_new = []
		for i in range(len(X)):
			X_new.append(np.transpose(X[i]))
		X = X_new
		self.X = X
		self.Y = Y
		# transpose to size: t x d x n

		self.task_num = len(X)
		self.dimension, _ = X[0].shape
		task_num = self.task_num
		dimension = self.dimension
		funcVal = []

		C0_prep = np.zeros(task_num)
		for t_idx in range(task_num):
			m1 = np.count_nonzero(Y[t_idx]==1)
			m2 = np.count_nonzero(Y[t_idx]==-1)
			if(m1==0 or m2==0):
				# inbalanced label
				C0_prep[t_idx] = 0
			else:
				C0_prep[t_idx] = np.log(m1/m2)

		if self.opts.init==2:
			W0 = np.zeros((dimension, task_num))
			C0 = np.zeros(task_num)
		elif self.opts.init == 0:
			W0 = np.random.randn(dimension, task_num)
			C0 = C0_prep
		else: 
			if hasattr(self.opts,'W0'):
				W0=self.opts.W0 
				if(W0.shape!=(dimension, task_num)):
					raise TypeError('\n Check input W0 size')
			else:
				W0 = np.zeros((dimension, task_num))
			if hasattr(self.opts, 'C0'):
				C0 = self.opts.C0
			else:
				C0 = C0_prep

		# this flag checks if gradient descent only makes significant step
		bFlag=0 

		Wz= W0
		Cz = C0
		Wz_old = W0
		Cz_old = C0

		t = 1
		t_old = 0

		gamma = 1
		gamma_inc = 2

		for it in trange(self.opts.maxIter, file=sys.stdout, desc='Training'):
			alpha = (t_old - 1)/t

			Ws = (1 + alpha) * Wz - alpha * Wz_old
			Cs = (1 + alpha) * Cz - alpha * Cz_old

			gWs, gCs, Fs = self.gradVal_eval(Ws, Cs)

			while True:
				Wzp = self.FGLasso_projection(Ws - gWs/gamma, self.rho1 / gamma)
				Czp = Cs - gCs/gamma
				Fzp = self.funVal_eval(Wzp, Czp)

				delta_Wzp = Wzp - Ws
				delta_Czp = Czp - Cs
				nrm_delta_Wzp = LA.norm(delta_Wzp)**2
				nrm_delta_Czp = LA.norm(delta_Czp)**2
				r_sum = (nrm_delta_Czp + nrm_delta_Wzp)/2

				Fzp_gamma = Fs + np.sum(delta_Wzp*gWs) + np.sum(delta_Czp*gCs)+ gamma/2 * r_sum*2
				if (r_sum <=1e-20):
					bFlag=1 # this shows that, the gradient step makes little improvement
					break
				if (Fzp <= Fzp_gamma):
					break
				else:
					gamma = gamma * gamma_inc
					
			Wz_old = Wz
			Cz_old = Cz
			Wz = Wzp
			Cz = Czp

			funcVal.append(Fzp + self.nonsmooth_eval(Wz, self.rho1))

			if (bFlag):
				print('\n The program terminates as the gradient step changes the solution very small.')
				break
			
			if(self.opts.tFlag == 0):
				if it>=2:
					if (abs( funcVal[-1] - funcVal[-2]) <= self.opts.tol):
						break
					
			elif(self.opts.tFlag == 1):
				if it>=2:
					if (abs( funcVal[-1] - funcVal[-2] ) <= self.opts.tol* funcVal[-2]):
						break
					
			elif(self.opts.tFlag == 2):
				if ( funcVal[-1]<= self.opts.tol):
					break
				
			elif(self.opts.tFlag == 3):
				if it>=self.opts.maxIter:
					break
			
			t_old = t
			t = 0.5 * (1 + (1+ 4 * t**2)**0.5)

		self.W = Wzp
		self.funcVal = funcVal

	def FGLasso_projection (self, D, lmbd):
		"""Lasso projection for panelties

		Args:
			D (np.array(np.array)): Weight matrix
			lmbd (int): panelties param

		Returns:
			(np.array(np.array)): panelties
		"""
		# l2.1 norm projection.
		ss = np.sum(D**2,axis=1)
		sq = np.sqrt(ss.astype(float))
		tmp = np.tile(np.maximum(0, 1 - lmbd/sq),(D.shape[1], 1))
		return np.transpose(tmp)*D

	# smooth part gradient.
	def gradVal_eval(self, W, C):
		"""Gradient Decent

		Args:
			W (np.array(np.array)): Weight Matrix with shape (d, t)
			C (np.array): intercept Matrix with shape (t, 1)

		Returns:
			grad_W (np.array(np.array)): gradient matrix of weight, shape (d, t)
		"""
		grad_W = np.zeros((self.dimension, self.task_num))
		grad_C = np.zeros(self.task_num)
		lossValVect = np.zeros((1, self.task_num)) 
		if self.opts.pFlag:
			# grad_W = zeros(zeros(W));
			# parfor i = 1:task_num
				# grad_W (i, :) = X{i}*(X{i}' * W(:,i)-Y{i})
			pass
		else:
			for i in range(self.task_num):
				grad_W[:,i], grad_C[i], lossValVect[:,i] = self.unit_grad_eval(W[:,i], C[i], i)
			grad_W = grad_W+ self.rho_L2 * 2 * W
			funcVal = np.sum(lossValVect) + self.rho_L2 * LA.norm(W)**2
			return grad_W, grad_C, funcVal

	def funVal_eval(self, W, C):
		"""Loss Accumulation

		Args:
			W ([np.array(np.array)]): weight matrix of shape (n, d, t)
			C ([np.array]): intercept Matrix with shape (t, n, 1)

		Returns:
			funcval (float): loss
		"""
		funcVal = 0
		if self.opts.pFlag:
			# parfor i = 1: task_num
			#     funcVal = funcVal + 0.5 * norm (Y{i} - X{i}' * W(:, i))^2;
			# end
			pass
		else:
			for i in range(self.task_num):
				funcVal = funcVal + self.unit_funcVal_eval(W[:,i], C[i], i)
			funcVal = funcVal + self.rho_L2 * LA.norm(W)**2
		return funcVal

	def nonsmooth_eval(self, W, rho_1):
		"""non-smooth loss evaluation

		Args:
			W (np.array(np.array)): weight matrix of shape (d, t)
			rho1 (float): L2,1-norm group Lasso parameter

		Returns:
			(float): loss 
		"""
		non_smooth_value = 0
		if self.opts.pFlag:
			pass
		else:
			for i in range(self.dimension):
				w = W[i, :]
				non_smooth_value = non_smooth_value+ rho_1 * LA.norm(w, ord=2)
			return non_smooth_value

	def unit_grad_eval(self, w, c, task_idx):
		"""Gradient decent in individual tasks

		Args:
			w (np.array): weight matrix of shape (d, 1), corresponding to individual task
			c (int): intercept Matrix with shape (1), corresponding to individual task
			task_idx (int): task index

		Returns:
			(np.array): gradient weight array
			(int): gradient intercept
			(int): task individual loss
		"""
		weight = np.ones((1, self.Y[task_idx].shape[0]))/self.task_num
		weighty = weight * self.Y[task_idx]
		z = -self.Y[task_idx]*(np.transpose(self.X[task_idx])@w + c)
		hinge = np.maximum(z, 0)
		funcVal = (weight @ (np.log(np.exp(-hinge)+np.exp(z-hinge))+hinge))
		prob = 1./(1+np.exp(z))
		z_prob = -weighty*(1-prob)
		grad_c = np.sum(z_prob)
		grad_w = self.X[task_idx]@np.transpose(z_prob)
		return grad_w.flatten(), grad_c, funcVal

	def unit_funcVal_eval(self, w, c, task_idx):
		"""individual loss in each task

		Args:
			w (np.array): weight matrix of shape (d, 1), corresponding to individual task
			c (int): intercept Matrix with shape (1), corresponding to individual task
			task_idx (int): task index

		Returns:
			(int): individual loss
		"""
		weight = np.ones((1, self.Y[task_idx].shape[0]))/self.task_num
		z = -self.Y[task_idx]*(np.transpose(self.X[task_idx])@w + c)
		hinge = np.maximum(z, 0)
		funcVal = (weight @ (np.log(np.exp(-hinge)+np.exp(z-hinge))+hinge))
		return funcVal

	def get_params(self, deep = False):
		"""Get inbult initalization params

		Args:
			deep (bool, optional): deep traverse. Defaults to False.

		Returns:
			(dict): dictionary of all inits
		"""
		return {'rho1':self.rho1, 'opts':self.opts}

	def _trained_parames(self):
		"""get all trained parameters

		Returns:
			(tuple): tuple containing: 
				([np.array(np.array)]): training weight matrix
				(float): final loss
		"""
		return self.W, self.funcVal

	def predict(self, X):
		"""Predict with test data

		Args:
			X [(np.array(np.array))]: input to predict, shape (t, n, d)

		Returns:
			([np.array()]): predict matrix, shape (t, n ,1)
		"""
		pred = []
		for i in range(self.task_num):
			pp = np.reshape(X[i], (-1, self.dimension)) @ self.W[:, i]
			pp = 1./(np.exp(-pp)+1)
			p = ((pp-0.5)>0)
			p = p+0
			pred.append(p.tolist())
		return pred






"""
.. module:: MTL_Least_L21
   :synopsis: MTL linear regression
.. moduleauthor:: Max J <https://github.com/DaPraxis>
"""
import numpy as np
from .init_opts import init_opts
from numpy import linalg as LA
from tqdm import tqdm
from tqdm import trange
import sys
import time

class MTL_Least_L21:
	""" MTL algorithm with least square regression and L21 penalty
	"""
	def __init__(self, opts, rho1=0.01):
		"""Initialization of MTL function

        Args:
            opts (opts): initalization class from opts
            rho1 (int, optional): L2,1-norm group Lasso parameter. Defaults to 0.01.
        """
		self.opts = init_opts(opts)
		self.rho1 = rho1
		self.rho_L2 = 0
		if hasattr(opts, 'rho_L2'):
			rho_L2 = opts.rho_L2

	def fit(self, X, Y, **kwargs):
		"""Fit with training samples and train
  
        t: task number
        
        n: number of entries
        
        d: data dimension

        Args:
            X ([np.array(np.array)]): t x n x d
            Y ([np.array(np.array)]): t x n x 1
        """
		if 'rho' in kwargs.keys():
			print(kwargs)
			self.rho1 = kwargs['rho']
		X_new = []
		for i in range(len(X)):
			X_new.append(np.transpose(X[i]))
		X = X_new
		self.X = X
		self.Y = Y
		# transpose to size: t x d x n

		self.task_num = len(X)
		self.dimension, _ = X[0].shape
		funcVal = []

		if hasattr(self.opts,'W0'):
			W0=self.opts.W0
		elif self.opts.init==2:
			W0 = np.zeros((self.dimension, self.task_num))
		elif self.opts.init == 0:
			W0 = np.zeros((self.dimension, self.task_num))
			for t_idx in range(self.task_num):
				W0[:,t_idx] = (X[t_idx]@Y[t_idx]).flatten()
		else:
			W0 = np.random.normal(0, 1, (self.dimension, self.task_num))

		# this flag checks if gradient descent only makes significant step
		bFlag=0 

		Wz= W0
		Wz_old = W0

		t = 1
		t_old = 0

		it = 0
		gamma = 1
		gamma_inc = 2

		for it in trange(self.opts.maxIter, file=sys.stdout, desc='Training'):
			alpha = (t_old - 1)/t
			Ws = (1 + alpha) * Wz - alpha * Wz_old
			# compute function value and gradients of the search point
			gWs = self.gradVal_eval(Ws)
			Fs = self.funVal_eval(Ws)
			in_it = 0

			while True:
				Wzp = self.FGLasso_projection(Ws - gWs/gamma, self.rho1 / gamma)
				Fzp = self.funVal_eval(Wzp)
				delta_Wzp = Wzp - Ws
				r_sum = LA.norm(delta_Wzp)**2
				Fzp_gamma = Fs + np.sum(delta_Wzp*gWs)+ gamma/2 * r_sum
				if (r_sum <=1e-20):
					bFlag=1 # this shows that, the gradient step makes little improvement
					break
				if (Fzp <= Fzp_gamma):
					break
				else:
					gamma = gamma * gamma_inc
					
			Wz_old = Wz
			Wz = Wzp
			funcVal.append(Fzp + self.nonsmooth_eval(Wz, self.rho1))

			if (bFlag):
				print('\n The program terminates as the gradient step changes the solution very small.')
				break
			
			if(self.opts.tFlag == 0):
				if it>=2:
					if (abs( funcVal[-1] - funcVal[-2]) <= self.opts.tol):
						break
					
			elif(self.opts.tFlag == 1):
				if it>=2:
					if (abs( funcVal[-1] - funcVal[-2] ) <= self.opts.tol* funcVal[-2]):
						break
					
			elif(self.opts.tFlag == 2):
				if ( funcVal[-1]<= self.opts.tol):
					break
				
			elif(self.opts.tFlag == 3):
				if it>=self.opts.maxIter:
					break
			
			t_old = t
			t = 0.5 * (1 + (1+ 4 * t**2)**0.5)

		self.W = Wzp
		self.funcVal = funcVal

	def FGLasso_projection (self, D, lmbd):
		"""Lasso projection for panelties

		Args:
			D (np.array(np.array)): Weight matrix
			lmbd (int): panelties param

		Returns:
			(np.array(np.array)): panelties
		"""
		# l2.1 norm projection.
		ss = np.sum(D**2,axis=1)
		sq = np.sqrt(ss.astype(float))
		tmp = np.tile(np.maximum(0, 1 - lmbd/sq),(D.shape[1], 1))
		return np.transpose(tmp)*D

	# smooth part gradient.
	def gradVal_eval(self, W):
		"""Gradient Decent

			Args:
				W (np.array(np.array)): Weight Matrix with shape (d, t)

			Returns:
				grad_W (np.array(np.array)): gradient matrix of weight, shape (d, t)
        """
		if self.opts.pFlag:
			# grad_W = zeros(zeros(W));
			# parfor i = 1:task_num
				# grad_W (i, :) = X{i}*(X{i}' * W(:,i)-Y{i})
			pass
		else:
			grad_W = []
			for i in range(self.task_num):
				g_w = np.reshape(self.X[i]@(np.reshape(np.transpose(self.X[i])@W[:,i], (self.Y[i].shape[0], 1))-np.reshape(self.Y[i], (self.Y[i].shape[0], 1))), (self.dimension,-1))
				grad_W.append(g_w)
			grad_W = np.concatenate(grad_W, axis=1)+ self.rho_L2 * 2 * W
			return grad_W

	def funVal_eval(self, W):
		"""Loss accumulation

        Args:
            W (np.array(np.array)): weight matrix of shape (d, t)

        Returns:
            funcval (float): loss
        """
		funcVal = 0
		if self.opts.pFlag:
			# parfor i = 1: task_num
			#     funcVal = funcVal + 0.5 * norm (Y{i} - X{i}' * W(:, i))^2;
			# end
			pass
		else:
			for i in range(self.task_num):
				funcVal = funcVal + 0.5 * LA.norm ((self.Y[i] - np.transpose(self.X[i]) @ W[:, i]), ord=2)**2
			funcVal = funcVal + self.rho_L2 * LA.norm(W)**2
		return funcVal

	def nonsmooth_eval(self, W, rho_1):
		"""non-smooth loss evaluation

		Args:
			W (np.array(np.array)): weight matrix of shape (d, t)
			rho1 (float): L2,1-norm group Lasso parameter

		Returns:
			(float): loss 
		"""
		non_smooth_value = 0
		if self.opts.pFlag:
			pass
		else:
			for i in range(W.shape[0]):
				w = W[i, :]
				non_smooth_value = non_smooth_value+ rho_1 * LA.norm(w, ord=2)
			return non_smooth_value

	def get_params(self, deep = False):
		"""Get inbult initalization params

		Args:
			deep (bool, optional): deep traverse. Defaults to False.

		Returns:
			(dict): dictionary of all inits
		"""
		return {'rho1':self.rho1, 'opts':self.opts}

	def predict(self, X):
		"""Predict with test data

		Args:
			X [(np.array(np.array))]: input to predict, shape (t, n, d)

		Returns:
			([np.array()]): predict matrix, shape (t, n ,1)
		"""
		pred = []
		for i in range(self.task_num):
			pred.append(np.reshape(X[i], (-1, self.dimension)) @ self.W[:, i])
		return pred






import sys
def init_opts(opts):
	# system floating point precision
	eps =  sys.float_info.epsilon
	"""
		opt is a opt object
	"""

	# Default values
	DEFAULT_MAX_ITERATION = 1000
	DEFAULT_TOLERANCE     = 1e-10
	MINIMUM_TOLERANCE     = eps * 100
	DEFAULT_TERMINATION_COND = 1
	DEFAULT_INIT = 0
	DEFAULT_PARALLEL_SWITCH = False

	# starting point
	if hasattr(opts, 'init'):
		if(opts.init!=0) and (opts.init!=1) and (opts.init!=2):
			opts.init=DEFAULT_INIT
		if (not hasattr(opts, 'W0')) and (opts.init==1):
			opts.init=DEFAULT_INIT
	else:
		opts.init = DEFAULT_INIT

	# tolerance
	if hasattr(opts, 'tol'):
		# detect if the tolerance is smaller than minimum tolerance allowed.
		if(opts.tol < MINIMUM_TOLERANCE):
			opts.tol = MINIMUM_TOLERANCE
	else:
		opts.tol = DEFAULT_TOLERANCE
    
	# Maximum iteration steps
	if hasattr(opts, 'maxIter'):
	    if (opts.maxIter < 1):
	        opts.maxIter = DEFAULT_MAX_ITERATION
	else:
	    opts.maxIter = DEFAULT_MAX_ITERATION

	# Termination condition
	if hasattr(opts, 'tFlag'):
	    if opts.tFlag < 0:
	        opts.tFlag = 0
	    elif opts.tFlag > 3:
	        opts.tFlag = 3
	    else:
	        opts.tFlag = int(opts.tFlag)
	else:
	    opts.tFlag = DEFAULT_TERMINATION_COND

	# Parallel Options: latter

	# if hasattr(opts, 'pFlag'):
	#     if opts.pFlag == true and not exist(mstring('matlabpool'), mstring('file')):
	#         opts.pFlag = false
	#         warning(mstring('MALSAR:PARALLEL'), mstring('Parallel Toolbox is not detected, MALSAR is forced to turn off pFlag.'))
	#     elif opts.pFlag != true and opts.pFlag != false:
	#         # validate the pFlag.
	#         opts.pFlag = DEFAULT_PARALLEL_SWITCH
	#     end
	# else:
	#     # if not set the pFlag to default.
	#     opts.pFlag = DEFAULT_PARALLEL_SWITCH
	# end

	# if opts.pFlag:
	#     # if the pFlag is checked,
	#     # check segmentation number.
	#     if hasattr(opts, 'pSeg_num'):
	#         if opts.pSeg_num < 0:
	#             opts.pSeg_num = matlabpool(mstring('size'))
	#         else:
	#             opts.pSeg_num = ceil(opts.pSeg_num)
	#         end
	#     else:
	#         opts.pSeg_num = matlabpool(mstring('size'))
	#     end
	# end

	return opts


from .test_data import get_data
from ..MTL_Cluster_Least_L21 import MTL_Cluster_Least_L21
from ..utils import MTL_data_extract, MTL_data_split, opts
import numpy as np
import math
from scipy import linalg
import plotly.express as px

# iris data
X_train, X_test, Y_train, Y_test, df = get_data()
opts = opts(1500,2)
opts.tol = 10**(-6)


# customized data
clus_var = 900
task_var = 16
nois_var = 150

clus_num = 2
clus_task_num = 10
task_num = clus_num * clus_task_num
sample_size = 100
dimension = 20
comm_dim = 2
clus_dim = math.floor((dimension - comm_dim)/2)

# generate cluster model
cluster_weight = np.random.randn(dimension, clus_num)* clus_var
for i in range(clus_num):
    bll = np.random.permutation(range(dimension-clus_num))<=clus_dim
    blc = np.array([False]*clus_num)
    bll = np.hstack((bll, blc))
    cluster_weight[:,i][bll]=0
cluster_weight[-1-comm_dim:, :]=0
W = np.tile(cluster_weight, (1, clus_task_num))
cluster_index = np.tile(range(clus_num), (1, clus_task_num)).T

# generate task and intra-cluster variance
W_it = np.random.randn(dimension, task_num) * task_var
for i in range(task_num):
    bll = np.hstack(((W[:-1-comm_dim+1,i]==0).reshape(1,-1), np.zeros((1,comm_dim))==1))
    W_it[:,i][bll.flatten()]=0
W = W+W_it

W = W + np.random.randn(dimension, task_num)*nois_var

X = [0]*task_num
Y = [0]*task_num
for i in range(task_num):
    X[i] = np.random.randn(sample_size, dimension)
    xw = X[i] @ W[:,i]
    s= xw.shape
    xw = xw + np.random.randn(s[0]) * nois_var
    Y[i] = np.sign(xw)
class Test_CMTL_Least_classification(object):
    """Pytest for Cluster Least Classification L21

    Args:
        object ([type]): entry point for pytest
    """
    
    def test_bsa_ihb(self):
        """ Test for bsa_ihb function inside CMTL usage
        """
        A = np.array([[1,2],[3,4]])
        EValue, EVector = linalg.eig(A) 
        Pz = np.real(EVector)
        # diag_EValue = np.real(np.diagonal(Evalue))
        diag_EValue = np.real(EValue).reshape((-1,1))
        clf = MTL_Cluster_Least_L21(opts, 3)
        x_star, t_star, it = clf.bsa_ihb(diag_EValue, 3)
        np.testing.assert_array_equal(x_star,
                              np.array([[0],[0]]))
        assert np.isnan(t_star) == True
        assert it == 3
    def test_basic_mat(self):
        """Test with basic matrix cases
        """
        clus_num = 2
        clus_task_num = 10
        task_num = clus_num * clus_task_num
        clf = MTL_Cluster_Least_L21(opts, 2)
        clf.fit(X, Y)
        corr = clf.analyse()
        # print(corr)
        # fig = px.imshow(corr, color_continuous_scale='Bluered_r')
        # fig.update_layout(
        # title={
        #     'text': "predict",
        #     })
        # fig.show()
        OrderedTrueModel = np.zeros(W.shape)
        clus_task_num = task_num//clus_num
        for i in range(clus_num):
            clusModel = W[:, i:task_num:clus_num]
            OrderedTrueModel[:, (i)*clus_task_num: (i+1)* clus_task_num] = clusModel
        corr2 = 1-np.corrcoef(OrderedTrueModel)
        # fig2 = px.imshow(corr2, color_continuous_scale='Bluered_r')
        # fig2.update_layout(
        # title={
        #     'text': "real",
        #     })
        # fig2.show()

    def test_check_simplified(self):
        """Test with simplified version matrix
        """
        # generate cluster model
        cluster_weight = np.ones((dimension, clus_num))* clus_var
        W = np.tile(cluster_weight, (1, clus_task_num))
        cluster_index = np.tile(range(clus_num), (1, clus_task_num)).T

        # generate task and intra-cluster variance
        W_it = np.zeros((dimension, task_num)) * task_var
        for i in range(task_num):
            bll = np.hstack(((W[:-1-comm_dim+1,i]==0).reshape(1,-1), np.zeros((1,comm_dim))==1))
            W_it[:,i][bll.flatten()]=0
        W = W+W_it

        W = W + np.zeros((dimension, task_num))*nois_var

        X = [0]*task_num
        Y = [0]*task_num
        for i in range(task_num):
            X[i] = np.ones((sample_size, dimension))
            xw = X[i] @ W[:,i]
            s= xw.shape
            xw = xw + np.ones((s[0])) * nois_var
            Y[i] = np.sign(xw)

        clf = MTL_Cluster_Least_L21(opts, 2)
        clf.fit(X, Y)
        corr = clf.analyse()
    # def test_iris_accuracy(self):
    #     clf = MTL_Cluster_Least_L21(opts, 3)
    #     clf.fit(X_train, Y_train)
    #     corr = clf.analyse()
    #     print(corr)


"""
.. module:: MTL_Cluster_Least_L21
   :synopsis: MTL Clustered linear regression
.. moduleauthor:: Max J <https://github.com/DaPraxis>
"""
import numpy as np
from .init_opts import init_opts
from numpy import linalg as LA
from tqdm import tqdm
from tqdm import trange
import sys
import time
from scipy.sparse import identity
from scipy import linalg
from scipy.sparse.linalg import spsolve
from scipy.sparse import isspmatrix

class MTL_Cluster_Least_L21:
    """Clustered MTL algorithm with least square regression and L21 penalty
    """
    def __init__(self, opts, k, rho1=10, rho2=0.1):
        """Initialization of CMTL function

        Args:
            opts (opts): initalization class from opts
            k (integer): number of clusters predefined
            rho1 (int, optional): [description]. Defaults to 10.
            rho2 (float, optional): [description]. Defaults to 0.1.
        """
        self.opts = init_opts(opts)
        self.rho1 = rho1
        self.rho2 = rho2
        self.rho_L2 = 0
        self.k = k
        if hasattr(opts, 'rho_L2'):
            rho_L2 = opts.rho_L2
    
    def fit(self, X, Y, **kwargs):
        """Fit with training samples and train
        t: task number
        n: number of entries
        d: data dimension

        Args:
            X ([np.array]): t x n x d
            Y ([np.array]): t x n x 1
        """
        if 'rho' in kwargs.keys():
            print(kwargs)
            self.rho1 = kwargs['rho']
        X_new = []
        for i in range(len(X)):
            X_new.append(np.transpose(X[i]))
        X = X_new
        self.X = X
        self.Y = Y
        # transpose to size: t x d x n
        self.task_num = len(X)
        self.dimension, _ = X[0].shape
        self.eta = self.rho2/self.rho1
        self.c = self.rho1 * self.eta * (1+self.eta)
        funcVal = []
        
        self.XY = [0]* self.task_num
        W0_prep = []
        for t in range(self.task_num):
            self.XY[t] = X[t] @ Y[t]
            W0_prep.append(self.XY[t].reshape((-1,1)))
        W0_prep = np.hstack(W0_prep)
        if hasattr(self.opts,'W0'):
            W0=self.opts.W0
        elif self.opts.init==2:
            W0 = np.zeros((self.dimension, self.task_num))
        elif self.opts.init == 0:
            W0 =W0_prep
        else:
            W0 = np.random.normal(0, 1, (self.dimension, self.task_num))
            
            
        M0 = np.array(identity(self.task_num)) * self.k / self.task_num
        # this flag checks if gradient descent only makes significant step
        
        bFlag=0 
        Wz= W0
        Wz_old = W0
        Mz = M0.toarray()
        Mz_old = M0.toarray()
          
        t = 1
        t_old = 0
        
        it = 0
        gamma = 1.0
        gamma_inc = 2

        for it in trange(self.opts.maxIter, file=sys.stdout, desc='Training'):
            alpha = (t_old - 1)/t
            Ws = (1 + alpha) * Wz - alpha * Wz_old
            if(isspmatrix(Mz)):
                Mz = Mz.toarray()
            if(isspmatrix(Mz_old)):
                Mz_old = Mz_old.toarray()
            Ms = (1 + alpha) * Mz - alpha * Mz_old
            # compute function value and gradients of the search point
            gWs, gMs, Fs = self.gradVal_eval(Ws, Ms)
            
            in_it = 0
            # for in_it in trange(2,file=sys.stdout, leave=False, unit_scale=True, desc='inner loop'):
            while True:
                Wzp = Ws - gWs/gamma
                Mzp, Mzp_Pz, Mzp_DiagSigz = self.singular_projection (Ms - gMs/gamma, self.k)
                Fzp = self.funVal_eval(Wzp, Mzp_Pz, Mzp_DiagSigz)
                
                delta_Wzs = Wzp - Ws
                delta_Mzs = Mzp - Ms
                
                r_sum = (LA.norm(delta_Wzs)**2 + LA.norm(delta_Mzs)**2)/2
                Fzp_gamma = Fs + np.sum(delta_Wzs*gWs) + np.sum(delta_Mzs*gMs) + gamma * r_sum
                if (r_sum <=1e-20):
                    bFlag=1 # this shows that, the gradient step makes little improvement
                    break
                if (Fzp <= Fzp_gamma):
                    break
                else:
                    gamma = gamma * gamma_inc
            Wz_old = Wz
            Wz = Wzp
            Mz_old = Mz
            Mz = Mzp
            funcVal.append(Fzp)
            
            if (bFlag):
                print('\n The program terminates as the gradient step changes the solution very small.')
                break
            if (self.opts.tFlag == 0):
                if it >= 2:
                    if (abs(funcVal[-1] - funcVal[-2]) <= self.opts.tol):
                        break
                    
            elif(self.opts.tFlag == 1):
                if it >= 2:
                    if (abs(funcVal[-1] - funcVal[-2]) <= self.opts.tol * funcVal[-2]):
                        break
                    
            elif(self.opts.tFlag == 2):
                if (funcVal[-1] <= self.opts.tol):
                    break
                
            elif(self.opts.tFlag == 3):
                if it >= self.opts.maxIter:
                    break
                
            t_old = t
            t = 0.5 * (1 + (1 + 4 * t ** 2) ** 0.5)
            
        self.W = Wzp
        self.M = Mzp
        self.funcVal = funcVal
    
    def singular_projection (self, Msp, k):
        """Projection of data

        Args:
            Msp (np.array(np.array)): M matrix with shape: (t, t)
            k (int): number of tasks

        Returns:
            (tuple): tuple containing:   
                W (np.array(np.array)): weight matrix of shape (d, t)
                M_Pz (np.array(np.array)): M matrix of shape (t, t)
                M_DiagSigz (np.array(np.array)): diagnolized M matrix optimized by bsa_ihb
        """
        # l2.1 norm projection.
        EValue, EVector = LA.eig(Msp) 
        idx = EValue.argsort()   
        EValue = EValue[idx]
        EVector = EVector[:,idx]
        Pz = np.real(EVector)
        diag_EValue = np.real(EValue)
        DiagSigz, _, _ = self.bsa_ihb(diag_EValue, k)
        Mzp = Pz @ np.diag(DiagSigz) @ Pz.T
        Mzp_Pz = Pz
        Mzp_DiagSigz = DiagSigz
        return Mzp, Mzp_Pz, Mzp_DiagSigz

    def bsa_ihb(self, eig_value, k):  
        """continuous quadratic knapsack problem solve in linear time
        
        Singular Projection
        
        min 1/2*||x - eig_value||_2^2
        
        s.t. b'\*x = k, 0<= x <= u,  b > 0

        Args:
            eig_value (np.array): eigenvalue of size (d, 1)
            k (int): number of clusters

        Returns:
            (tuple): tuple containing: 
                x_star (np.array): optimized solution with Newton's Method, shape (d, 1)
                t_star (float): intercepts
                it (int): iteration
        """   
        
        break_flag = 0
        b = np.ones(eig_value.shape)
        u = np.ones(eig_value.shape)
        
        t_l = eig_value/b
        t_u = (eig_value - u)/b
        T = np.concatenate((t_l, t_u), axis=0)
        t_L = -np.Infinity
        t_U = np.Infinity
        g_tL = 0.
        g_tU = 0.
        
        it = 0
        while(len(T)!=0):
            it +=1
            g_t = 0.
            t_hat = np.median(T)
            
            U = t_hat < t_u
            M = (t_u <= t_hat) & (t_hat <= t_l)
            
            if np.sum(U):
                g_t += np.sum(b[U]*u[U])
            if np.sum(M):
                g_t += np.sum(b[M]*(eig_value[M]-t_hat*b[M]))
            if g_t > k:
                t_L = t_hat
                T = T[T>t_hat]
                g_tL = g_t
            elif g_t <k:
                t_U = t_hat
                T = T[T<t_hat]
                g_tU = g_t
            else:
                t_star = t_hat
                break_flag = 1
                break
        if not break_flag:
            eps = g_tU - g_tL
            t_star = t_L - (g_tL - k) * (t_U - t_L)/(eps)
        est = eig_value-t_star * b
        if(np.isnan(est).any()):
            est[np.isnan(est)] = 0
        x_star = np.minimum(u, np.max(est, 0))
        return x_star, t_star, it
    
    def gradVal_eval(self, W, M):
        """Gradient Decent

        Args:
            W (np.array(np.array)): Weight Matrix with shape (d, t)
            M (np.array(np.array)): M matrix shape (t, t)

        Returns:
            (tuple): tuple containing: 
                grad_W (np.array(np.array)): gradient matrix of weight, shape (d, t)
                grad_M (np.array(np.array)): gradient matrix of M, shape (t, t)
                funcval (float): loss
        """
        IM = self.eta * identity(self.task_num)+M
        # could be sparse matrix to solve
        invEtaMWt = linalg.inv(IM) @ W.T
        if self.opts.pFlag:
            # grad_W = zeros(zeros(W));
            # # parfor i = 1:task_num
            # # grad_W (i, :) = X{i}*(X{i}' * W(:,i)-Y{i})
            pass
        else:
            grad_W = []
            for i in range(self.task_num):
                XWi = self.X[i].T @ W[:,i]
                XTXWi = self.X[i] @ XWi
                grad_W.append((XTXWi - self.XY[i]).reshape(-1,1))
            grad_W = np.hstack(grad_W)
            grad_W = grad_W + 2 * self.c * invEtaMWt.T
            W2 = W.T @ W
            grad_M = - self.c * W2@linalg.inv(IM)@linalg.inv(IM)
        funcVal = 0
        if self.opts.pFlag:
            pass
        else:
            for i in range(self.task_num):
                funcVal = funcVal + 0.5 * LA.norm ((self.Y[i] - self.X[i].T @ W[:, i]), ord=2)**2
        funcVal = funcVal + self.c * np.trace( W @ invEtaMWt)
        return grad_W, grad_M, funcVal
			
    def funVal_eval(self, W, M_Pz, M_DiagSigz):
        """Loss accumulation

        Args:
            W (np.array(np.array)): weight matrix of shape (d, t)
            M_Pz (np.array(np.array)): M matrix of shape (t, t)
            M_DiagSigz (np.array(np.array)): diagnolized M matrix optimized by bsa_ihb

        Returns:
            (float): loss
        """
        invIM = M_Pz @ (np.diag(1/(self.eta + np.array(M_DiagSigz)))) @ M_Pz.T
        invEtaMWt = invIM @ W.T
        funcVal = 0
        if self.opts.pFlag:
            # parfor i = 1: task_num
            # #     funcVal = funcVal + 0.5 * norm (Y{i} - X{i}' * W(:, i))^2;
            # # end
            pass
        else:
            for i in range(self.task_num):
                funcVal = funcVal + 0.5 * LA.norm ((self.Y[i] - self.X[i].T @ W[:, i]), ord=2)**2
            funcVal = funcVal + self.c * np.trace(W @ invEtaMWt)
        return funcVal

    def get_params(self, deep = False):
        """Get inbult initalization params

        Args:
            deep (bool, optional): deep traverse. Defaults to False.

        Returns:
            (dict): dictionary of all inits
        """
        return {'rho1':self.rho1, 'rho2':self.rho2,'opts':self.opts, 'k':self.k}

    def get_weights(self):
        """Get weight matrix

        Returns:
            (np.array(np.array)): Weight matrix
        """
        return self.W
    
    def analyse(self):
        """Analyse weight matrix cross clusters with correlation

        Returns:
            (np.array(np.array)): correlation map
        """
        # returns correlation matrix
  
        kmCMTL_OrderedModel = np.zeros(self.W.shape)
        clus_task_num = self.task_num//self.k
        for i in range(self.k):
            clusModel = self.W[:, i:self.task_num:self.k]
            kmCMTL_OrderedModel[:, (i)*clus_task_num: (i+1)* clus_task_num] = clusModel
        return 1-np.corrcoef(kmCMTL_OrderedModel)
     






from sklearn.model_selection import train_test_split
import numpy as np
import pandas as pd
from collections import defaultdict, OrderedDict
import plotly.figure_factory as ff
from sklearn.cluster import DBSCAN

class opts:
	def __init__(self, maxIter, init):
		self.maxIter = maxIter
		self.init = init
		self.pFlag = False
  

def MTL_data_split(X, Y, test_size=0.4, random_state=0):
	"""
		X: shape: t x n x d
		Y: shape: t x n x 1
	"""
	t = len(X)
	X_train = []
	X_test = []
	y_train = []
	y_test = []
	for i in range(t):
		X_train_s, X_test_s, y_train_s, y_test_s = train_test_split(X[i], Y[i], test_size=test_size, random_state=random_state)
		X_train.append(X_train_s)
		X_test.append(X_test_s)
		y_train.append(y_train_s.flatten())
		y_test.append(y_test_s.flatten())
	return X_train, X_test, y_train, y_test

def MTL_ClusterBoosted_data_extract_auto(df, target_feat, clf = DBSCAN()):
    db = clf.fit(df.values)
    df['t_target'] = db.labels_
    return MTL_data_extract(df, 't_target', target_feat)

def MTL_data_extract(df, task_feat, target):
	'''
		df: extract from data frame
		task_feat: feature categorized as task
		target: feature that serves as target

		ret:
			X
			Y
	'''
	tasks = df[task_feat].unique()
	Y = []
	X = []
	for t in tasks:
		tmp1 = df.loc[df[task_feat]==t]
		tmp2 = tmp1.loc[:, tmp1.columns != task_feat]
		tmp = tmp2.loc[:, tmp2.columns != target]
		x = tmp.values
		X.append(np.array(x))
		y = tmp1.loc[:, df.columns == target].values
		Y.append(np.array(y))
	return X, Y

def RFA(df, tasks, mtl_clf, target, top=10):
    def reformat(cols, w, top=10):
        RFA = OrderedDict()
        cols = np.array(cols)
        fet, task = w.shape
        total = {}
        all_tasks = []
        for i in range(task):
            col = w[:,i].flatten()
            index = sorted(range(len(col)), key=lambda i: col[i], reverse=True)[:top]
            e = set(cols[index])
            RFA["task {}".format(i+1)] = e
            all_tasks.append("task {}".format(i+1))
            total = set.union(e, total)
        print("all top {} colns are {}".format(top, total))
        ret = defaultdict(lambda: [])
        df_v = pd.DataFrame(False, index=list(total), columns=all_tasks)
        df_v2 = pd.DataFrame(None, index = list(total), columns=[str(p+1) for p in range(len(all_tasks))])
        df_RFA = []
        for t in all_tasks:
            df_RFA.append(list(RFA[t]))
        for i in total:
            count = 1
            for k, v in RFA.items():
                if i in v:
                    ret[i].append(k)
                    df_v[k][i]=True
                    df_v2[str(count)][i] = int(k[-2:])
                    if(len(k)==6):
                        df_v2[str(count)][i] = int(k[-1])
                    count+=1
        return df_v, all_tasks, list(total), df_v2, df_RFA, RFA
        
    def sort_df(df):
        fet, tsk = df.values.shape
        ret = pd.DataFrame(None, columns=list(df.columns))
        ind = list(df.index)
        seq = []
        for i in range(tsk):
            for j in range(fet):
                if(np.count_nonzero(~np.isnan(list(df_v2.iloc[j].values)))==i+1):
                    ret.loc[len(ret)] = df_v2.iloc[j].values
                    seq.append(ind[j])
        ret = ret.rename(index={i:j for i,j in zip(range(fet), seq)})
        return ret
    
    def get_z_text(z, mp):
        x, y = z.shape
        ret = np.empty([x, y],dtype="S10")
        for i in range(x):
            for j in range(y):
                ret[i][j]=mp[z[i][j]]
            return ret.astype(str).tolist()
        
    all_col = (df.loc[:, (df.columns != target)&(df.columns != tasks)].columns).tolist()
    df_v, all_tasks, total, df_v2, RFA, index = reformat(all_col, mtl_clf.W, top=top)
    fet, task = mtl_clf.W.shape
    mp = {i+1:"Task_{}".format(i) for i in range(task)}
    mp[None] = ''
    mp[np.nan] = ''
    df_v3 = sort_df(df_v2)
    z_text = get_z_text(df_v3.values, mp)
    fig = ff.create_annotated_heatmap(z = df_v3.values.tolist(), annotation_text=z_text, y=list(df_v3.index))
    fig.update_xaxes(showticklabels=False, showgrid=False)
    return fig
	



import numpy as np
import pandas as pd
from sklearn import datasets
from ..utils import MTL_data_extract, MTL_data_split, opts, MTL_ClusterBoosted_data_extract_auto

class Test_Data(object):
    def get_data(self):
        iris = datasets.load_iris()
        df = pd.DataFrame(data= np.c_[iris['data'], iris['target']],
                            columns= iris['feature_names'] + ['target'])
        df['cat1']=0
        df['cat2']=0
        df['target'] = df['target'].astype(int)
        df.loc[df['petal width (cm)']<=0.8, 'cat1'] = 0
        df.loc[(df['petal width (cm)']>0.8) & (df['petal width (cm)']<=1.6), 'cat1'] = 1
        df.loc[(df['petal width (cm)']>1.6) & (df['petal width (cm)']<=2.4), 'cat1'] = 2
        df.loc[df['petal length (cm)']<=2.3, 'cat2'] = 0
        df.loc[(df['petal length (cm)']>2.3) & (df['petal length (cm)']<=4.6), 'cat2'] = 1
        df.loc[(df['petal length (cm)']>4.6) & (df['petal length (cm)']<=6.9), 'cat2'] = 2
        # X_i, Y_i = MTL_data_extract(df, 'cat2', 'target')
        X_i, Y_i = MTL_ClusterBoosted_data_extract_auto(df, 'cat2')
        X_train, X_test, Y_train, Y_test = MTL_data_split(X_i, Y_i, test_size=0.4)
        return X_train, X_test, Y_train, Y_test, df


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="MD-MTL",
    version="0.0.9",
    author="Max Jiang",
    author_email="haoyanhy.jiang@mail.utoronto.ca",
    description="An Ensemble Med-Multi-Task Learning Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Interactive-Media-Lab-Data-Science-Team/Vampyr-MTL",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from ..MTL_Softmax_L21 import MTL_Softmax_L21
import numpy as np
import pandas as pd
from sklearn import datasets
from sklearn import preprocessing
from ..utils import MTL_data_extract, MTL_data_split, opts
from .test_data import get_data
from sklearn.linear_model import LogisticRegression
import os

opts = opts(1000,2)

X_train, X_test, Y_train, Y_test, df = get_data()

df2 = df.copy()
df2.loc[df['target']==0, 'target'] = 'flower1'
df2.loc[df['target']==1, 'target'] = 'flower2'
df2.loc[df['target']==2, 'target'] = 'flower3'
X_i, Y_i = MTL_data_extract(df2, 'cat2', 'target')
X_train_c, X_test_c, Y_train_c, Y_test_c = MTL_data_split(X_i, Y_i, test_size=0.4)

print(os.getcwd())
print('???????????????')
# df3 = pd.read_csv('./cleaned_BRFSS.csv')

def normalize(X):
    for i in range(len(X)):
        min_max_scaler = preprocessing.MinMaxScaler()
        X[i] = min_max_scaler.fit_transform(X[i])
    return X

class Test_softmax_classification(object):
    # def test_real_data(self):
    #     df4 = df3[(df3['ADDEPEV2']==2)|(df3['ADDEPEV2']==1)]
    #     # opts.tol = 1e-20
    #     X, Y = MTL_data_extract(df4, "ADDEPEV2", "_BMI5CAT")
    #     task = [0]*2
    #     taskT = 0
    #     for i in range(1):
    #         X_train, X_test, Y_train, Y_test = MTL_data_split(X, Y, test_size=0.998)
    #         X_train = normalize(X_train)
    #         X_test = normalize(X_test)
    #         for i in range(len(Y_train)):
    #             Y_train[i] = Y_train[i].astype(int)
    #         clf = MTL_Softmax_L21(opts)
    #         clf.fit(X_train, Y_train)
    #         pred = clf.predict(X_test)
            
    #         c_t = 0
    #         total = 0
    #         for i in range(len(pred)):
    #             correct = np.sum(pred[i]==Y_test[i])
    #             sub = len(pred[i])
    #             task[i] = max(task[i], correct/sub*100)
    #             total += sub
    #             c_t += correct
    #         taskT = max(taskT, c_t/total*100)
    #     print("accurcy for task 1 is {}%".format(task[0]))
    #     print("accurcy for task 2 is {}%".format(task[1]))
    #     print("total accuracy is {}%".format(taskT))
        
    #     for i in range(len(pred)):
    #         clf = LogisticRegression(random_state=0).fit(X_train[i], Y_train[i])
    #         s = clf.score(X_test[i], Y_test[i])
    #         print("SKLearn accuracy for task {} is {}%".format(i, s*100))
    
    #     assert c_t/total*100 == 0
    
    def test_soft_numerical_accuracy(self):
        ult_thres = 0.5
        thres = 0.9
        its = 10
        succ = 0
        for it in range(its):
            clf = MTL_Softmax_L21(opts)
            clf.fit(X_train, Y_train, rho=0.00001)
            pred = clf.predict(X_test)
            correct = 0
            total = 0
            for i, j in zip(Y_test, pred):
                for k,l in zip(i,j):
                    if(k == l):
                        correct+=1
                    total+=1
            acc = correct/total
            if(acc>thres):
                succ+=1    
                print("pass with accuracy {}".format(acc))
            else:
                print("fail with accuracy {}".format(acc))
                pass
        print(">>>>>>>>>>>>>>>>>>>>>")
        print("in total of {} tests, {} passes {} threshold".format(its, acc, thres))
        print(">>>>>>>>>>>>>>>>>>>>>")
        assert succ/its >=ult_thres
    
    def test_soft_cat_accuracy(self):
        ult_thres = 0.5
        thres = 0.9
        its = 10
        succ = 0
        for it in range(its):
            clf = MTL_Softmax_L21(opts)
            clf.fit(X_train_c, Y_train_c, rho=0.00001)
            acc = clf.score(X_test_c, Y_test_c)
            if(acc>thres):
                succ+=1    
                # print("pass with accuracy {}".format(acc))
            else:
                # print("fail with accuracy {}".format(acc))
                pass
        print(">>>>>>>>>>>>>>>>>>>>>")
        print("in total of {} tests, {} passes {} threshold".format(its, succ/its, thres))
        print(">>>>>>>>>>>>>>>>>>>>>")
        assert succ/its >=ult_thres
